#include <stdio.h>
#include <spurs/spsa/spsa.h>
#include <spurs/club/suf.h>
#include <spu_mfcio.h>
#include <spu_intrinsics.h>
#include "data.h"

#define TIMER_COEFF 16.66667
#define MCL_DMA_TAG_ID 0

// YUV画像のバッファ
uint32_t buf_y[0x400] __attribute__((aligned(128)));
uint16_t buf_u[0x400] __attribute__((aligned(128)));
uint16_t buf_v[0x400] __attribute__((aligned(128)));


// 最内ループのマクロ
#define INNERLOOP() {                                   \
    vb = spu_msub(spu_mul(vf2, va), vb, vy);            \
    va = spu_sub(va2, spu_add(vb2, vx));                \
    va2 = spu_mul(va, va);                              \
    vb2 = spu_mul(vb, vb);                              \
    vtmp = spu_add(va2, vb2);                           \
    flag = spu_cmpgt(vf4, vtmp);                        \
    vcount = spu_sub(vcount, spu_and(vui1, flag));      \
    flagui = spu_extract(spu_gather(flag), 0);          \
    if (flagui == 0) break;}


// ローカルメモリから読み込んで転送完了を待つ
// ls:LS側アドレス　ea:ローカルメモリ側実効アドレス　size:転送サイズ
void mcl_load_and_wait(void * ls, unsigned int ea, unsigned int size)
{
  unsigned int mask;

  // 読み込みDMA実行
  mfc_get(ls, ea, size, MCL_DMA_TAG_ID, 0, 0);

  // DMAが完了するまで待つ
  mask = mfc_read_tag_mask();
  mfc_write_tag_mask(1 << MCL_DMA_TAG_ID);
  mfc_write_tag_update_all();
  mfc_read_tag_status();
  mfc_write_tag_mask(mask);
}


// ローカルメモリに書き込んで転送完了を待つ
// ls:LS側アドレス　ea:ローカルメモリ側実効アドレス　size:転送サイズ
void mcl_save_and_wait(void * ls, unsigned int ea, unsigned int size)
{
  unsigned int mask;

  // 書き込みDMA実行
  mfc_put(ls, ea, size, MCL_DMA_TAG_ID, 0, 0);

  // DMAが完了するまで待つ
  mask = mfc_read_tag_mask();
  mfc_write_tag_mask(1 << MCL_DMA_TAG_ID);
  mfc_write_tag_update_all();
  mfc_read_tag_status();
  mfc_write_tag_mask(mask);
}


// マイクロ秒単位のスリープ
void mcl_usleep(unsigned int time)
{
  unsigned int count = (unsigned int)((float)time * TIMER_COEFF);

  // イベントマスクを保存
  uint32_t mask_bak = spu_read_event_mask();

  // イベントをリセット
  spu_write_event_mask(0);
  spu_write_event_ack(~0);

  // SPU Decrementerにタイマー値をセット
  spu_writech(SPU_WrDec, count);

  // デクリメンタイベントをセット
  spu_write_event_mask(MFC_DECREMENTER_EVENT);

  // デクリメンタが0になるまでスリープ
  uint32_t events = spu_read_event_status();
  spu_write_event_ack(events);

  // イベントマスクを復元
  spu_write_event_mask(mask_bak);
}


// マンデルブロ集合を描画
void job(mandel_params_t * param)
{
  // vector unsigned intの下位1バイトをunsigned intにまとめるためのシャッフルパターン
  vector unsigned char vsp_vtou = {3,7,11,15,128,128,128,128,128,128,128,128,128,128,128,128};
  vector unsigned char vsp_vtos = {128,128,3,11,128,128,128,128,128,128,128,128,128,128,128,128};
  vector unsigned int vui1 = spu_splats(1u);
  vector unsigned int vmaxcount;
  vector float vf2 = spu_splats(2.0f);
  vector float vf4 = spu_splats(4.0f);
  unsigned int i, j, k, xres, yres, vsize, hsize, page, startline, endline, addr_ea_y, addr_ea_u, addr_ea_v, para_my, para_all, frame, max_count, out_ea;
  float xmin, xmax, ymin, ymax, dx, dy, zoom, cx, cy, color_u, color_v;
  // LS上のピクセルデータのポインタ（4バイト単位）
  uint32_t *pix;
  void *ptmp;
  uint16_t *pixu;
  void *ptmpu;
  uint16_t *pixv;
  void *ptmpv;

  xres = param->xres;
  yres = param->yres;
  cx = param->cx;
  cy = param->cy;
  zoom = param->zoom;
  frame = param->frame;
  max_count = param->max_count;
  color_u = param->color_u;
  color_v = param->color_v;
  out_ea = param->out_ea;
  para_my = param->para_my;
  para_all = param->para_all;
  hsize = spurs::YUV_HEADER_WIDTH_HD_I420;
  vsize = spurs::YUV_HEADER_HEIGHT_HD_I420;

  vmaxcount = spu_splats((unsigned int)max_count);

  xmin = cx - zoom * (float)xres / (float)yres;
  xmax = cx + zoom * (float)xres / (float)yres;
  ymin = cy - zoom;
  ymax = cy + zoom;

  dx = (xmax - xmin) / (float)xres;
  dy = (ymax - ymin) / (float)yres;
  startline = para_my;
  endline = yres;

  vector float vxmin = spu_splats(xmin);
  vxmin = spu_insert(xmin + dx, vxmin, 1);
  vxmin = spu_insert(xmin + dx * 2, vxmin, 2);
  vxmin = spu_insert(xmin + dx * 3, vxmin, 3);
  vector float vymin = spu_splats(ymin);
  vector float vdx = spu_splats(dx * 4);
  vector float vdy = spu_splats(dy);

  // バッファをクリア
  for (i = 0; i < hsize / 2; i++)
  {
    vector unsigned int * p;
    p = (vector unsigned int *)(void *)buf_y;
    *p = spu_splats(0x10101010u);
    p += hsize / 32;
    *p = spu_splats(0x10101010u);
    p = (vector unsigned int *)(void *)buf_u;
    *p = spu_splats(0x80808080u);
    p = (vector unsigned int *)(void *)buf_v;
    *p = spu_splats(0x80808080u);
  }

  // ダブルバッファのページをリセット
  page = 0;

  for (i = startline; i < endline; i += para_all)
  {
    // 偶数ライン:0 奇数ライン:1
    unsigned int even = i & 1;

    // 書き込み先アドレス
    addr_ea_y = out_ea + sizeof(spurs::yuv_header_t) + hsize * i ;
    addr_ea_u = out_ea + sizeof(spurs::yuv_header_t) + hsize * vsize + (hsize / 2) * (i / 2);
    addr_ea_v = addr_ea_u + hsize * vsize / 4;

    vector float vy = spu_add(vymin, spu_mul(vdy, spu_convtf(spu_splats(i), 0)));

    // ピクセルバッファのポインタをLS上の汎用バッファの先頭にセット
    pix = buf_y;
    pixu = buf_u;
    pixv = buf_v;

    // ダブルバッファのためのオフセットをプラス
    pix += page * 512;
    pixu += page * 512;
    pixv += page * 512;

    ptmp = pix;
    ptmpu = pixu;
    ptmpv = pixv;

    for (j = 0; j < xres / 4; j++)
    {
      vector float vtmp;
      vector unsigned int flag;
      vector float vx = spu_add(vxmin, spu_mul(vdx, spu_convtf(spu_splats(j), 0)));
      vector float va = vx;
      vector float vb = vy;
      vector float va2 = spu_mul(va, va);
      vector float vb2 = spu_mul(vb, vb);
      vector unsigned int vcount = vmaxcount;

      for (k = 0; k < max_count; k+=8)
      {
        unsigned int flagui;
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
        INNERLOOP();
      }

      // LS上のピクセルバッファにピクセルの値を書き込む
      vector float vcountf = spu_convtf(vcount, 0);
      vector unsigned int vcounty = spu_convtu(spu_mul(vcountf, spu_splats(24556.8f / ((float)max_count * (float)max_count))), 0);
      *pix = spu_extract(spu_shuffle(spu_splats(235u) - vcounty, spu_splats(0u), vsp_vtou), 0);

      // 奇数ラインの時はUVの計算を省略
      if (even == 0)
      {
        // y:16-235 u,v:16-240
        vector unsigned int vcountu = spu_convtu(spu_splats(128.0f) + vcountf * spu_splats(color_u / (float)max_count), 0);
        vector unsigned int vcountv = spu_convtu(spu_splats(128.0f) + vcountf * spu_splats(color_v / (float)max_count), 0);
        *pixu = (unsigned short)spu_extract(spu_shuffle(vcountu, spu_splats(0u), vsp_vtos), 0);
        *pixv = (unsigned short)spu_extract(spu_shuffle(vcountv, spu_splats(0u), vsp_vtos), 0);
      }

      // LS上のピクセルバッファのポインタを進める
      pix++;
      pixu++;
      pixv++;
    }

    // ダブルバッファの前回のDMAが完了するまで待つ
    mfc_write_tag_mask(1 << page);
    mfc_write_tag_update_all();
    mfc_read_tag_status();

    // ローカルメモリのフレームバッファに書き込み
    // 書き込みDMA実行。転送サイズは16バイト単位に補正
    mfc_put(ptmp, addr_ea_y, hsize & ~15, page, 0, 0);
    // 奇数ラインの時はUVの書き込みを省略
    if (even == 0)
    {
      mfc_put(ptmpu, addr_ea_u, (hsize / 2) & ~15, page, 0, 0);
      mfc_put(ptmpv, addr_ea_v, (hsize / 2) & ~15, page, 0, 0);
    }

    // ダブルバッファを切り替え(0と1で交互に)
    page = page ^ 1;
  }

  // 最後のDMAが完了するまで待つ
  mfc_write_tag_mask(1 << page);
  mfc_write_tag_update_all();
  mfc_read_tag_status();
}

